@@ -1,18 +1,30 @@
 #include <cstdlib>
 #include <cstring>
+#include <fcntl.h>
 #include <filesystem>
 #include <fstream>
-
-#include "utils.h"
-#include "gmock/gmock.h"
-#include "gtest/gtest.h"
-#include <fcntl.h>
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <unistd.h>
 
+#include "util/bpf_names.h"
+#include "util/cgroup.h"
+#include "util/io.h"
+#include "util/kernel.h"
+#include "util/math.h"
+#include "util/paths.h"
+#include "util/similar.h"
+#include "util/strings.h"
+#include "util/symbols.h"
+#include "util/system.h"
+#include "util/wildcard.h"
+#include "gmock/gmock-matchers.h"
+#include "gtest/gtest.h"
+
 namespace bpftrace::test::utils {
 
+using namespace bpftrace::util;
+
 TEST(utils, split_string)
 {
   std::vector<std::string> tokens_empty = {};
@@ -150,13 +162,13 @@ static std::string get_working_path()
     throw std::runtime_error(
         "getting current working directory for tests failed");
   }
-  return std::string(cwd_path);
+  return cwd_path;
 }
 
 TEST(utils, resolve_binary_path)
 {
   std::string path = "/tmp/bpftrace-test-utils-XXXXXX";
-  if (::mkdtemp(&path[0]) == nullptr) {
+  if (::mkdtemp(path.data()) == nullptr) {
     throw std::runtime_error("creating temporary path for tests failed");
   }
 
@@ -191,7 +203,7 @@ TEST(utils, abs_path)
 {
   std::string path = "/tmp/bpftrace-test-utils-XXXXXX";
   std::string rel_file = "bpftrace-test-utils-abs-path";
-  if (::mkdtemp(&path[0]) == nullptr) {
+  if (::mkdtemp(path.data()) == nullptr) {
     throw std::runtime_error("creating temporary path for tests failed");
   }
 
@@ -236,7 +248,7 @@ TEST(utils, get_cgroup_path_in_hierarchy)
 {
   std::string tmpdir = "/tmp/bpftrace-test-utils-XXXXXX";
 
-  if (::mkdtemp(&tmpdir[0]) == nullptr) {
+  if (::mkdtemp(tmpdir.data()) == nullptr) {
     throw std::runtime_error("creating temporary path for tests failed");
   }
 
@@ -314,20 +326,77 @@ static void with_env(const std::string &key,
                      const std::string &val,
                      std::function<void()> fn)
 {
+  char *old_val_ptr = ::getenv(key.c_str());
+  std::string old_val = {};
+
+  // we must capture the existing key's value by value, since the invokd
+  // function might modify the environment and invalidate old_val_ptr.
+  if (old_val_ptr != nullptr) {
+    old_val = old_val_ptr;
+  }
+
+  auto restore_key = [&key, &old_val]() {
+    if (!old_val.empty()) {
+      EXPECT_EQ(::setenv(key.c_str(), old_val.c_str(), 1), 0);
+    } else {
+      EXPECT_EQ(::unsetenv(key.c_str()), 0);
+    }
+  };
+
   EXPECT_EQ(::setenv(key.c_str(), val.c_str(), 1), 0);
+
   try {
     fn();
   } catch (const std::exception &ex) {
-    EXPECT_EQ(::unsetenv(key.c_str()), 0);
+    restore_key();
     throw ex;
   }
-  EXPECT_EQ(::unsetenv(key.c_str()), 0);
+
+  restore_key();
+}
+
+TEST(utils, with_env_nonexisting_key)
+{
+  // Test nonexisting variable
+  const std::string nonexisting_key = "nonexisting_key";
+  const std::string some_value = "some_value";
+
+  with_env(nonexisting_key, some_value, [&]() {
+    EXPECT_EQ(::getenv(nonexisting_key.c_str()), some_value);
+  });
+
+  EXPECT_EQ(::getenv(nonexisting_key.c_str()), nullptr);
+}
+
+TEST(utils, with_env_restoration)
+{
+  // Test that an existing variable is restored correctly,
+  // even with nested environment mutation
+  const std::string existing_key = "existing_key";
+  const std::string existing_value = "existing_value";
+  const std::string random_new_key = "random_new_key";
+  const std::string random_new_value = "random_new_value";
+  const std::string some_value = "some_value";
+
+  ::setenv(existing_key.c_str(), existing_value.c_str(), 1);
+
+  with_env(existing_key, some_value, [&]() {
+    EXPECT_EQ(::getenv(existing_key.c_str()), some_value);
+    ::unsetenv(existing_key.c_str());
+    ::setenv(random_new_key.c_str(), random_new_value.c_str(), 1);
+  });
+
+  EXPECT_EQ(::getenv(existing_key.c_str()), existing_value);
+
+  // Cleanup
+  ::unsetenv(existing_key.c_str());
+  ::unsetenv(random_new_key.c_str());
 }
 
 TEST(utils, find_in_path)
 {
   std::string tmpdir = "/tmp/bpftrace-test-utils-XXXXXX";
-  ASSERT_TRUE(::mkdtemp(&tmpdir[0]));
+  ASSERT_TRUE(::mkdtemp(tmpdir.data()));
 
   // Create some directories
   const std::filesystem::path path(tmpdir);
@@ -391,11 +460,11 @@ TEST(utils, find_in_path)
 // Hopefully they are easy to maintain. If not, please delete.
 TEST(utils, find_near_self)
 {
-  auto runtime_tests = find_near_self("runtime-tests.sh");
+  auto runtime_tests = find_near_self("near_self_file");
   // clang-tidy is not aware ASSERT_*() terminates testcase
   // NOLINTBEGIN(bugprone-unchecked-optional-access)
   ASSERT_TRUE(runtime_tests.has_value());
-  EXPECT_TRUE(runtime_tests->filename() == "runtime-tests.sh");
+  EXPECT_TRUE(runtime_tests->filename() == "near_self_file");
   EXPECT_TRUE(std::filesystem::exists(*runtime_tests));
   // NOLINTEND(bugprone-unchecked-optional-access)
 
@@ -405,12 +474,11 @@ TEST(utils, find_near_self)
 TEST(utils, get_pids_for_program)
 {
   auto pids = get_pids_for_program("/proc/self/exe");
-
-  ASSERT_EQ(pids.size(), 1);
-  ASSERT_EQ(pids[0], getpid());
+  ASSERT_TRUE(bool(pids));
+  EXPECT_THAT(*pids, testing::Contains(getpid()));
 
   pids = get_pids_for_program("/doesnotexist");
-  ASSERT_EQ(pids.size(), 0);
+  ASSERT_FALSE(bool(pids));
 }
 
 TEST(utils, round_up_to_next_power_of_two)
@@ -427,4 +495,63 @@ TEST(utils, round_up_to_next_power_of_two)
   ASSERT_EQ(round_up_to_next_power_of_two(max_power_of_two), max_power_of_two);
 }
 
+TEST(utils, cat_file_success)
+{
+  std::string test_content = "Hello, cat_file test!\nThis is line 2.\n";
+  char filename[] = "/tmp/bpftrace-test-cat-file-XXXXXX";
+  int fd = mkstemp(filename);
+  ASSERT_NE(fd, -1) << "Failed to create temporary file";
+  ASSERT_EQ(write(fd, test_content.c_str(), test_content.length()),
+            static_cast<ssize_t>(test_content.length()));
+  close(fd);
+
+  // Test cat_file with the temporary file
+  std::stringstream out;
+  cat_file(filename, 1024, out);
+
+  // Verify output matches the file content
+  EXPECT_EQ(test_content, out.str());
+
+  // Cleanup
+  unlink(filename);
+}
+
+TEST(utils, cat_file_nonexistent)
+{
+  // Path to a file that shouldn't exist
+  std::string nonexistent_file = "/tmp/bpftrace-nonexistent-file-test-XXXXXX";
+  int fd = mkstemp(const_cast<char *>(nonexistent_file.c_str()));
+  close(fd);
+  unlink(nonexistent_file.c_str()); // Ensure file doesn't exist
+
+  testing::internal::CaptureStderr();
+
+  // Test cat_file with nonexistent file
+  std::stringstream out;
+  cat_file(nonexistent_file.c_str(), 1024, out);
+
+  // Get captured stderr
+  std::string stderr_output = testing::internal::GetCapturedStderr();
+
+  // Verify no output was produced
+  EXPECT_TRUE(out.str().empty())
+      << "cat_file should not output anything for nonexistent files";
+
+  // Verify error message was logged
+  EXPECT_THAT(stderr_output, testing::HasSubstr("failed to open file"))
+      << "Error message should indicate file opening failure";
+}
+
+TEST(utils, similar)
+{
+  // This is not well-defined, and therefore we cannot include whitebox tests
+  // that go into much detail. These tests provide a generic `very different`
+  // versus `quite similar` sanity baseline.
+  EXPECT_FALSE(is_similar("foo", "bar"));
+  EXPECT_FALSE(is_similar("foo", "baz"));
+  EXPECT_TRUE(is_similar("foo", "foofoo"));
+  EXPECT_TRUE(is_similar("fo", "foo"));
+  EXPECT_TRUE(is_similar("foobar", "fobar"));
+}
+
 } // namespace bpftrace::test::utils
